# -*- coding: utf-8 -*-
"""
Adapter code for pyftpdlib
"""

from __future__ import annotations

from typing import Callable, Iterable, Optional, Tuple
import logging
import os
import stat
import time
from collections import namedtuple
from io import BytesIO

from pyftpdlib.authorizers import AuthenticationFailed
from pyftpdlib.filesystems import AbstractedFS
from pyftpdlib.handlers import FTPHandler as _FTPHandler

from commonutil_net_fileservice.config import (DEFAULT_REV_CONTENT, DEFAULT_REV_FILENAME, User, make_users_map)

_log = logging.getLogger(__name__)

_ALL_PERM = "elradfmwM"
"""
Read permissions:
- "e" = change directory (CWD command)
- "l" = list files (LIST, NLST, STAT, MLSD, MLST, SIZE, MDTM commands)
- "r" = retrieve file from the server (RETR command)
Write permissions:
- "a" = append data to an existing file (APPE command)
- "d" = delete file or directory (DELE, RMD commands)
- "f" = rename file or directory (RNFR, RNTO commands)
- "m" = create directory (MKD command)
- "w" = store a file to the server (STOR, STOU commands)
- "M" = change file mode (SITE CHMOD command)
- "T" = update file last modified time (MFMT command)
"""

_StatResult = namedtuple("_StatResult", (
		"st_mode",
		"st_ino",
		"st_dev",
		"st_nlink",
		"st_uid",
		"st_gid",
		"st_size",
		"st_atime",
		"st_mtime",
		"st_ctime",
))
_BOOTUP_TSTAMP = time.time()


class Authorizer:
	__slots__ = (
			'users',
			'base_folder_path',
	)

	def __init__(self, users: Iterable[User], base_folder_path: str, *args, **kwds):
		super().__init__(*args, **kwds)
		self.users = make_users_map(users)
		self.base_folder_path = os.path.abspath(base_folder_path)

	def update_users(self, users: Iterable[User]):
		rec = make_users_map(users)
		to_drop = []
		for k in self.users:
			if k not in rec:
				to_drop.append(k)
		for k in to_drop:
			self.users.pop(k, None)
		for k, u in rec.items():
			self.users[k] = u

	def validate_authentication(self, username, password, handler):  # pylint: disable=unused-argument
		"""
		Raises AuthenticationFailed if supplied username and password don't
		match the stored credentials, else return None.
		"""
		u = self.users.get(username)
		if not u:
			raise AuthenticationFailed(f"user not found: {username!r}")
		if not u.check_credential(password):
			raise AuthenticationFailed(f"authorize failed: {username!r}")

	def get_home_dir(self, username):
		"""
		Return the user's home directory.
		Since this is called during authentication (PASS), AuthenticationFailed
		can be freely raised by subclasses in case the provided username no
		longer exists.
		"""
		u = self.users[username]
		wd = os.path.abspath(os.path.join(self.base_folder_path, u.username))
		if not wd.startswith(self.base_folder_path):
			raise AuthenticationFailed(f"home folder escaped: {username!r}: {wd!r}")
		return wd

	def impersonate_user(self, username, password):
		"""
		Impersonate another user (noop).
		It is always called before accessing the filesystem.
		By default it does nothing.  The subclass overriding this method is
		expected to provide a mechanism to change the current user.
		"""

	def terminate_impersonation(self, username):
		"""
		Terminate impersonation (noop).
		It is always called after having accessed the filesystem. By default it
		does nothing.  The subclass overriding this method is expected to
		provide a mechanism to switch back to the original user.
		"""

	def has_user(self, username):
		"""
		Whether the username exists in the virtual users table.
		"""
		return username in self.users

	def has_perm(self, username, perm, path=None):  # pylint: disable=unused-argument
		"""
		Whether the user has permission over path (an absolute pathname of a
		file or a directory).
		Expected perm argument is one of the following letters: "elradfmwMT".
		"""
		return perm in _ALL_PERM

	def get_perms(self, username):  # pylint: disable=unused-argument
		"""
		Return current user permissions.
		"""
		return _ALL_PERM

	def get_msg_login(self, username):
		"""
		Return the user's login message.
		"""
		return f"Login successful ({username})"

	def get_msg_quit(self, username):
		"""
		Return the user's quitting message.
		"""
		return f"Bye ({username})"


# FileSystem = AbstractedFS


class NamedBytesIO(BytesIO):
	def __init__(self, name, *args, **kwds):
		super().__init__(*args, **kwds)
		self.name = name


class FileSystem(AbstractedFS):
	_v_rev_filename = DEFAULT_REV_FILENAME
	_v_rev_content = DEFAULT_REV_CONTENT.encode('utf-8', 'ignore')
	_v_rev_stat = _StatResult(stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH, 0, 0, 1, 0, 0, len(_v_rev_content), _BOOTUP_TSTAMP, _BOOTUP_TSTAMP, _BOOTUP_TSTAMP)

	def __init__(self, root, cmd_channel, *args, **kwds):
		super().__init__(root, cmd_channel, *args, **kwds)
		self._rev_path = os.path.normpath(os.path.join(self.root, self._v_rev_filename))
		self._begin_relpath = self._index_relpath_begin()

	@classmethod
	def set_v_rev_file(cls, filename: str, content: str) -> None:
		content = content.strip() + "\n"
		cls._v_rev_filename = filename
		cls._v_rev_content = content.encode('utf-8', 'ignore')
		cls._v_rev_stat = _StatResult(stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH, 0, 0, 1, 0, 0, len(cls._v_rev_content), _BOOTUP_TSTAMP, _BOOTUP_TSTAMP,
										_BOOTUP_TSTAMP)

	def _index_relpath_begin(self):
		p = self.root
		if p[-1] == "/":
			return len(p)
		return (len(p) + 1)

	def fs2relpath(self, filepath):
		try:
			return filepath[self._begin_relpath:]
		except Exception:
			pass
		return None

	def open(self, filename, mode):
		if filename == self._rev_path:
			return NamedBytesIO(self._v_rev_filename, self._v_rev_content)
		return super().open(filename, mode)

	def listdir(self, path):
		"""
		List the content of a directory.
		"""
		_log.debug("listdir: %r", path)
		result = super().listdir(path)
		if path == self._root:
			return result + [
					self._v_rev_filename,
			]
		return result

	def listdirinfo(self, path):
		"""
		List the content of a directory.
		"""
		_log.debug("listdirinfo: %r", path)
		return self.listdir(path)

	def stat(self, path):
		"""
		Perform a stat() system call on the given path.
		"""
		_log.debug("stat: %r", path)
		if self._rev_path == path:
			return self._v_rev_stat
		return super().stat(path)

	def lstat(self, path):
		"""
		Perform a lstat() system call on the given path.
		"""
		_log.debug("lstat: %r", path)
		if self._rev_path == path:
			return self._v_rev_stat
		return super().lstat(path)

	def isfile(self, path):
		"""
		Return True if path is a file.
		"""
		_log.debug("isfile: %r", path)
		if self._rev_path == path:
			return True
		return super().isfile(path)

	def getsize(self, path):
		"""
		Return the size of the specified file in bytes.
		"""
		_log.debug("getsize: %r", path)
		if self._rev_path == path:
			return len(self._v_rev_content)
		return super().getsize(path)

	def getmtime(self, path):
		"""
		Return the last modified time as a number of seconds since the epoch.
		"""
		_log.debug("getmtime: %r", path)
		if self._rev_path == path:
			return _BOOTUP_TSTAMP
		return super().getmtime(path)


class FTPHandler(_FTPHandler):
	process_callable = None
	default_transfer_type = 'i'

	def on_connect(self):
		_log.info("connect: (%r:%r), (%r connections/ %r total connections)", self.remote_ip, self.remote_port, self.server.ip_map.count(self.remote_ip),
					len(self.server.ip_map))

	def on_login(self, username):
		"""
		Called on user login.
		"""
		self._current_type = self.default_transfer_type
		_log.info("user login: %r (%r:%r)", username, self.remote_ip, self.remote_port)

	def on_login_failed(self, username, password):
		"""
		Called on failed login attempt.
		At this point client might have already been disconnected if it failed
		too many times.
		"""
		_log.info("login rejected: %r (%r:%r)", username, self.remote_ip, self.remote_port)

	def on_logout(self, username):
		"""
		Called when user "cleanly" logs out due to QUIT or USER issued twice (re-login).
		This is not called if the connection is simply closed by client.
		"""
		_log.info("user logout: %r (%r:%r)", username, self.remote_ip, self.remote_port)

	def on_disconnect(self):
		"""
		Called when connection is closed.
		"""
		self.file_rule_set = None
		_log.info("disconnect: %r (%r:%r), (%r connections/ %r total connections)", self.username, self.remote_ip, self.remote_port,
					self.server.ip_map.count(self.remote_ip), len(self.server.ip_map))

	def on_file_sent(self, file):
		"""
		Called every time a file has been succesfully sent.
		`file` is the absolute name of the file just being sent.
		"""
		_log.info("sent: (by %r) %r", self.username, file)

	def on_file_received(self, file):
		"""
		Called every time a file has been succesfully received.
		`file` is the absolute name of the file just being received.
		"""
		_log.info("recv: (by %r) %r", self.username, file)
		if not self.fs:
			return
		relfilepath = self.fs.fs2relpath(file)
		_log.debug("got %r from remote", relfilepath)
		if not relfilepath:
			return
		if not FTPHandler.process_callable:
			return
		try:
			remote_location = f"{self.remote_ip}:{self.remote_port}"
		except Exception as e:
			remote_location = f"UNKNOWN: {e!r}"
			_log.exception("cannot have remote location (user=%r)", self.username)
		try:
			FTPHandler.process_callable(self.username, remote_location, file, relfilepath)  # pylint: disable=not-callable
		except Exception:
			_log.exception("caught exception on invoking received file processor: user=%r, remote=%r, filename=%r, relfilepath=%r", self.username,
							remote_location, file, relfilepath)

	def ftp_USER(self, line):
		"""
		Set the username for the current session.
		"""
		try:
			line = line.lower()
		except Exception as e:
			self.respond(f"530 Invalid username: {line!r} ({e!r}).")
			return
		_FTPHandler.ftp_USER(self, line)

	def ftp_RMD(self, path):
		"""
		Remove the specified directory.
		On success return the directory path, else None.
		"""
		_log.info("rmdir: (by %r) %r", self.username, path)
		return _FTPHandler.ftp_RMD(self, path)

	def ftp_DELE(self, path):
		"""
		Delete the specified file.
		On success return the file path, else None.
		"""
		_log.info("dele: (by %r) %r", self.username, path)
		return _FTPHandler.ftp_DELE(self, path)

	def ftp_RNFR(self, path):
		"""
		Rename the specified (only the source name is specified here, see RNTO command)
		"""
		_log.info("rnfr: (by %r) %r", self.username, path)
		return _FTPHandler.ftp_RNFR(self, path)

	def ftp_RNTO(self, path):
		"""
		Rename file (destination name only, source is specified with RNFR).
		On success return a (source_path, destination_path) tuple.
		"""
		_log.info("rnto-request: (by %r) %r", self.username, path)
		ret = _FTPHandler.ftp_RNTO(self, path)
		if ret is not None:
			_log.info("rnto: (by %r) %r", self.username, ret)
		return ret


# pylint: disable=too-many-arguments
def setup_handlers(banner_text: str,
					base_folder_path: str,
					users: Iterable[User],
					process_callable: Optional[Callable],
					ftp_passive_port_range: Optional[Iterable[int]],
					ftp_public_address: Optional[str],
					v_rev_filename: str = DEFAULT_REV_FILENAME,
					v_rev_content: str = DEFAULT_REV_CONTENT) -> Tuple[FTPHandler, Authorizer]:
	"""
	Prepare handlers for FTP server
	"""
	ftp_hnd = FTPHandler
	auth_hnd = Authorizer(users, base_folder_path)
	ftp_hnd.authorizer = auth_hnd
	fsabs = FileSystem  # implementation default
	fsabs.set_v_rev_file(v_rev_filename, v_rev_content)
	ftp_hnd.abstracted_fs = fsabs
	ftp_hnd.banner = banner_text
	ftp_hnd.passive_ports = ftp_passive_port_range
	if ftp_public_address:
		ftp_hnd.masquerade_address = ftp_public_address
	ftp_hnd.permit_foreign_addresses = True
	ftp_hnd.process_callable = process_callable
	return (ftp_hnd, auth_hnd)



# -*- coding: utf-8 -*-
"""
Adapter code for paramiko
"""

from __future__ import annotations

from typing import Any, Callable, Iterable, List
import _thread
import logging
import os
import subprocess

import paramiko

from commonutil_net_fileservice.scp import SCPSinkHandler

_log = logging.getLogger(__name__)


def stream_exec_stdin(channel: paramiko.Channel, pobj: subprocess.Popen):
	loaded_bytes = 0
	streamed_bytes = 0
	try:
		while True:
			buf = channel.recv(2048)
			s = len(buf)
			if s == 0:
				pobj.stdin.close()
				_log.debug("stdin streamed %d/%d bytes", streamed_bytes, loaded_bytes)
				return
			loaded_bytes = loaded_bytes + s
			pobj.stdin.write(buf)
			streamed_bytes = streamed_bytes + s
	except Exception:
		_log.exception("cannot stream stdin data (streamed %d/%d bytes)", streamed_bytes, loaded_bytes)
		pobj.stdin.close()


def stream_exec_stdout(channel: paramiko.Channel, pobj: subprocess.Popen):
	loaded_bytes = 0
	streamed_bytes = 0
	try:
		while True:
			buf = pobj.stdout.read(2048)
			s = len(buf)
			if s == 0:
				_log.debug("stdout streamed %d/%d bytes", streamed_bytes, loaded_bytes)
				return
			loaded_bytes = loaded_bytes + s
			channel.sendall(buf)
			streamed_bytes = streamed_bytes + s
	except Exception:
		_log.exception("cannot stream stdout data (streamed %d/%d bytes)", streamed_bytes, loaded_bytes)


def stream_exec_stderr(channel: paramiko.Channel, pobj: subprocess.Popen):
	loaded_bytes = 0
	streamed_bytes = 0
	try:
		while True:
			buf = pobj.stderr.read(2048)
			s = len(buf)
			if s == 0:
				return
			loaded_bytes = loaded_bytes + s
			channel.sendall_stderr(buf)
			streamed_bytes = streamed_bytes + s
	except Exception:
		_log.exception("cannot stream stderr data (streamed %d/%d bytes)", streamed_bytes, loaded_bytes)


def rewrite_target_path(user_folder_path: str, target_path: str) -> str:
	if (not target_path) or (target_path == '.'):
		return user_folder_path
	end_with_slash = (target_path[-1] == '/')
	result_path = os.path.abspath(os.path.join(user_folder_path, target_path.strip('/\\')))
	if not result_path.startswith(user_folder_path):
		result_path = user_folder_path
	if end_with_slash:
		result_path = result_path + '/'
	return result_path


def run_rsync_exec(user_folder_path: str, _report_callable: Callable, channel: paramiko.Channel, cmdpart: List[str], binpath_rsync: str):
	cmdpart[-1] = rewrite_target_path(user_folder_path, cmdpart[-1])
	_log.debug('rsync command (rewritten): %r', cmdpart)
	with subprocess.Popen(cmdpart, bufsize=0, executable=binpath_rsync, stdin=subprocess.PIPE, stdout=subprocess.PIPE, close_fds=True) as pobj:
		_thread.start_new_thread(stream_exec_stdout, (channel, pobj))
		_thread.start_new_thread(stream_exec_stdin, (channel, pobj))
		# _thread.start_new_thread(stream_exec_stderr, (channel, pobj))  # , stderr=subprocess.PIPE
		_log.debug('stream ready')
		retcode = pobj.wait()
	_log.debug('command stopped: %r', retcode)
	# TODO: scan for changes
	channel.send_exit_status(retcode)
	channel.close()


class _SCPResponseCallable:
	__slots__ = ('channel', )

	def __init__(self, channel: paramiko.Channel) -> None:
		self.channel = channel

	def __call__(self, is_success: bool, message_text: str, *args: Any, **kwds: Any) -> None:
		if is_success:
			self.channel.sendall(b'\x00')
			return
		self.channel.sendall(b'\x01' + message_text.encode(encoding='utf-8') + b"\n")


def run_scp_sink(user_folder_path: str, report_callable: Callable, channel: paramiko.Channel, cmdpart: Iterable):
	_log.debug('scp command: %r', cmdpart)
	target_path = rewrite_target_path(user_folder_path, cmdpart[-1])
	streamed_bytes = 0
	loaded_bytes = 0
	resp_fn = _SCPResponseCallable(channel)
	scp_sink = SCPSinkHandler(user_folder_path, target_path, report_callable=report_callable)
	try:
		resp_fn(True, '')
		while True:
			buf = channel.recv(2048)
			s = len(buf)
			if s == 0:
				scp_sink.close()
				_log.debug("scp streamed %d/%d bytes", streamed_bytes, loaded_bytes)
				channel.send_exit_status(0)
				break
			loaded_bytes = loaded_bytes + s
			scp_sink.feed(buf, resp_fn)
			streamed_bytes = streamed_bytes + s
	except Exception:
		_log.exception("cannot stream stdin data (streamed %d/%d bytes)", streamed_bytes, loaded_bytes)
		channel.send_exit_status(1)
	finally:
		scp_sink.close()
	channel.close()



# -*- coding: utf-8 -*-
"""
Configuration data classes
"""

from __future__ import annotations

from typing import Any, Callable, Iterable, Mapping, Optional
import logging
import os

_log = logging.getLogger(__name__)

_REV = "0.0.2; 54e6001c024d5e934f47fe6f4451084986e8f9f0"  # REV-CONSTANT:full 5d022db7d38f580a850cd995e26a6c2f

DEFAULT_REV_FILENAME = "_rev-info.txt"
DEFAULT_REV_CONTENT = _REV + "\n"


def _default_credential_checker(u: User, remote_credential: str) -> bool:
	return (u.credential == remote_credential)


class SSHPKey:
	__slots__ = (
			'key_type',
			'b64_text',
	)

	def __init__(self, key_type: str, b64_text: str) -> None:
		self.key_type = key_type
		self.b64_text = b64_text


def unpack_ssh_pkey(pkey_text: str) -> Optional[SSHPKey]:
	aux = pkey_text.split(' ', 3)
	if len(aux) < 2:
		return None
	return SSHPKey(aux[0], aux[1])


class User:
	__slots__ = (
			'username',
			'prebuild_folders',
			'credential',
			'ssh_pkeys',
	)

	credential_checker: Callable[[User, str], bool] = _default_credential_checker

	def __init__(self, username: str, prebuild_folders: Optional[Iterable[str]], credential: Any, ssh_pkeys: Optional[Iterable[SSHPKey]]) -> None:
		self.username = username
		self.prebuild_folders = prebuild_folders if prebuild_folders else ()
		self.credential = credential
		self.ssh_pkeys = ssh_pkeys if ssh_pkeys else ()

	def check_credential(self, remote_credential: str) -> bool:
		""" Return True if given `remote_credential` is accepted.
		"""
		return self.credential_checker(remote_credential)

	def check_ssh_pkey(self, key_type: str, b64_text: str) -> Optional[SSHPKey]:
		""" Return SSHPKey instance if matching key is found.
		"""
		for pk in self.ssh_pkeys:
			if (pk.key_type == key_type) and (pk.b64_text == b64_text):
				return pk
		return None

	def prepare_user_folders(self, base_folder_path: str) -> None:
		user_folder_path = os.path.abspath(os.path.join(base_folder_path, self.username))
		if not self.prebuild_folders:
			os.makedirs(user_folder_path, exist_ok=True)
			return
		for d_path in self.prebuild_folders:
			target_path = os.path.abspath(os.path.join(user_folder_path, d_path))
			if not target_path.startswith(user_folder_path):
				_log.warning("escaped user pre-build folder (username=%r, user-folder=%r): %r", self.username, user_folder_path, target_path)
				continue
			os.makedirs(target_path, exist_ok=True)


def make_users_map(users: Iterable[User]) -> Mapping[str, User]:
	result = {}
	for u in users:
		result[u.username] = u
	return result



# -*- coding: utf-8 -*-
"""
SCP protocol handler
"""

from __future__ import annotations

from typing import Callable, Optional, Tuple
import logging
import os

_log = logging.getLogger(__name__)


class SCPSinkHandler:
	__slots__ = (
			'_user_folder_path',
			'_work_folder_path',
			'_mkdirmode',
			'_createfilemode',
			'_report_callable',
			'_parser',
			'_filesize',
			'_filename',
			'_dirstack',
			'_dirname',
			'_recv_fp',
			'_recv_done',
			'_last_absfilepath',
			'_last_relfilepath',
	)

	# pylint: disable=too-many-arguments
	def __init__(self,
					user_folder_path: str,
					work_folder_path: str,
					mkdirmode: int = 0o755,
					createfilemode: int = 0o644,
					report_callable: Optional[Callable[[str, str], None]] = None) -> None:
		self._user_folder_path = user_folder_path
		self._work_folder_path = os.path.abspath(work_folder_path)
		self._mkdirmode = mkdirmode
		self._createfilemode = createfilemode
		self._report_callable = report_callable
		self._parser = self._start
		self._filesize = 0
		self._filename = []
		self._dirstack = []
		self._dirname = []
		self._recv_fp = None
		self._recv_done = 0
		self._last_absfilepath = ''
		self._last_relfilepath = ''

	def feed(self, d: bytes, resp_fn: Callable[[bool, str], None]):
		while len(d) > 0:
			remain_d, next_parser = self._parser(d, resp_fn)
			if next_parser is not None:
				self._parser = next_parser
			d = remain_d

	def _close_recv_fp(self):
		if self._recv_fp is None:
			return
		try:
			self._recv_fp.close()
		except Exception:
			_log.exception("close receving file failed")
		self._recv_fp = None
		try:
			if self._report_callable:
				self._report_callable(self._last_absfilepath, self._last_relfilepath)
		except Exception:
			_log.exception("invoke report callable with (%r, %r) failed", self._last_absfilepath, self._last_relfilepath)

	def close(self):
		self._close_recv_fp()

	def _prepare_dirstack_folder(self):
		p = os.path.abspath(os.path.join(self._work_folder_path, *self._dirstack))
		if not p.startswith(self._user_folder_path):
			_log.warning("escaped dir stack (%r): %r", self._user_folder_path, p)
			return
		os.makedirs(p, self._mkdirmode, exist_ok=True)

	def _open_file_path(self) -> Optional[str]:
		n = bytes(self._filename).decode(encoding='utf-8').strip('/\\')
		if not n:
			return None
		pathcomps = self._dirstack + [
				n,
		]
		p = os.path.abspath(os.path.join(self._work_folder_path, *pathcomps))
		if not p.startswith(self._user_folder_path):
			_log.warning("escaped file name (%r): %r", self._user_folder_path, p)
			return None
		fp = open(p, 'wb', opener=self._fp_opener)  # pylint: disable=consider-using-with
		self._last_absfilepath = p
		self._last_relfilepath = p[len(self._user_folder_path) + 1:]
		return fp

	def _fp_opener(self, path, flags):
		return os.open(path, flags, self._createfilemode)

	def _start(self, d: bytes, resp_fn: Callable) -> Tuple[bytes, Callable]:
		cmdch = d[0]
		if cmdch == 0x43:  # b'C':
			self._filesize = 0
			self._filename = []
			self._recv_fp = None
			self._recv_done = 0
			return (d[1:], self._c_mode0)
		if cmdch == 0x54:  # b'T':
			return (d[1:], self._t_drop)
		if cmdch == 0x44:  # b'D':
			self._dirname = []
			return (d[1:], self._d_mode0)
		if cmdch == 0x45:  # b'E':
			return (d[1:], self._e_popdir)
		resp_fn(False, f"unknown command character {cmdch!r}")
		_log.error('unknown command character: %r', cmdch)
		raise Exception('unknown command character', cmdch)

	def _c_mode0(self, d: bytes, _resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		for idx, ch in enumerate(d):
			if ch == 0x20:  # b' ':
				return (d[idx:], self._c_mode1)
		return (b'', None)

	def _c_mode1(self, d: bytes, _resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		for idx, ch in enumerate(d):
			if ch != 0x20:  # b' ':
				return (d[idx:], self._c_size0)
		return (b'', None)

	def _c_size0(self, d: bytes, _resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		for idx, ch in enumerate(d):
			if ch in b'0123456789':
				self._filesize = (self._filesize * 10) + (ch - 0x30)  # 0x30 = b'0'
			elif ch == 0x20:  # b' ':
				return (d[idx:], self._c_size1)
		return (b'', None)

	def _c_size1(self, d: bytes, _resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		for idx, ch in enumerate(d):
			if ch != 0x20:  # ' ':
				return (d[idx:], self._c_filename)
		return (b'', None)

	def _c_filename(self, d: bytes, resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		for idx, ch in enumerate(d):
			if ch == 0x0A:  # b"\n":
				resp_fn(True, '')
				self._recv_fp = self._open_file_path()
				return (d[idx + 1:], self._c_file_recv0)
			self._filename.append(ch)
		return (b'', None)

	def _c_file_recv0(self, d: bytes, _resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		buf_size = len(d)
		remain_size = self._filesize - self._recv_done
		if remain_size == 0:
			f_content = b''
		elif buf_size > remain_size:
			f_content = d[:remain_size]
			d = d[remain_size:]
		else:
			f_content = d
			d = b''
		if self._recv_fp:
			self._recv_fp.write(f_content)
		self._recv_done = self._recv_done + len(f_content)
		if self._filesize == self._recv_done:
			self._close_recv_fp()
			self._recv_fp = None
			return (d, self._c_file_recv1)
		return (d, None)

	def _c_file_recv1(self, d: bytes, resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		ch = d[0]
		if ch == 0x00:  # b'\x00':
			resp_fn(True, '')
		else:
			resp_fn(False, f"unexpected stop character {ch!r}")
			raise Exception('unexpected stop character', ch)
		return (d[1:], self._start)

	def _t_drop(self, d: bytes, resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		for idx, ch in enumerate(d):
			if ch == 0x0A:  # b"\n":
				resp_fn(True, '')
				# self._recv_fp = None # TODO: save timestamp for later use
				return (d[idx + 1:], self._start)
		return (b'', None)

	def _d_mode0(self, d: bytes, _resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		for idx, ch in enumerate(d):
			if ch == 0x20:  # b' ':
				return (d[idx:], self._d_mode1)
		return (b'', None)

	def _d_mode1(self, d: bytes, _resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		for idx, ch in enumerate(d):
			if ch != 0x20:  # b' ':
				return (d[idx:], self._d_size0)
		return (b'', None)

	def _d_size0(self, d: bytes, _resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		for idx, ch in enumerate(d):
			if ch == 0x20:  # b' ':
				return (d[idx:], self._d_size1)
		return (b'', None)

	def _d_size1(self, d: bytes, _resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		for idx, ch in enumerate(d):
			if ch != 0x20:  # ' ':
				return (d[idx:], self._d_dirname)
		return (b'', None)

	def _d_dirname(self, d: bytes, resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		for idx, ch in enumerate(d):
			if ch == 0x0A:  # b"\n":
				n = bytes(self._dirname).decode(encoding='utf-8').strip('/\\')
				self._dirstack.append(n)
				self._prepare_dirstack_folder()
				resp_fn(True, '')
				return (d[idx + 1:], self._start)
			self._dirname.append(ch)
		return (b'', None)

	def _e_popdir(self, d: bytes, resp_fn: Callable[[bool, str], None]) -> Tuple[bytes, Callable]:
		for idx, ch in enumerate(d):
			if ch == 0x0A:  # b"\n":
				if len(self._dirstack) > 0:
					self._dirstack.pop()
					# TODO: setup other folder variables
				resp_fn(True, '')
				return (d[idx + 1:], self._start)
		return (b'', None)



# -*- coding: utf-8 -*-
"""
Adapter code for paramiko
"""

from __future__ import annotations

from typing import Callable, Iterable, Optional, Tuple
import _thread
import io
import logging
import os
import shutil
import socketserver
import time

import paramiko

from commonutil_net_fileservice.config import (DEFAULT_REV_CONTENT, DEFAULT_REV_FILENAME, User, make_users_map)
from commonutil_net_fileservice.paramikosubexec import (run_rsync_exec, run_scp_sink)

_log = logging.getLogger(__name__)


def open_mode_from_flags(flags) -> str:
	if flags & os.O_EXCL:
		return 'xb'
	if flags & os.O_WRONLY:
		return 'ab' if (flags & os.O_APPEND) else 'wb'
	if flags & os.O_RDWR:
		return 'a+b' if (flags & os.O_APPEND) else 'r+b'
	return 'rb'


def _fp_opener(path, flags):
	return os.open(path, flags, 0o644)


def _remote_location_from_transport(transport: paramiko.Transport) -> str:
	try:
		remote_addr, remote_port = transport.getpeername()
		return f"{remote_addr}:{remote_port}"
	except Exception as e:
		return f"exception:{e!r}"


def make_sftp_attr(file_name: str, file_content: bytes) -> paramiko.SFTPAttributes:
	fattr = paramiko.SFTPAttributes()
	fattr.st_size = len(file_content)
	fattr.st_uid = 0
	fattr.st_gid = 0
	fattr.st_mode = 0o0444
	fattr.filename = file_name
	return fattr


class SFTPHandle(paramiko.SFTPHandle):
	# pylint: disable=too-many-arguments
	def __init__(self, fp, flags, local_path, rel_path, report_callable, stat_object=None):
		super().__init__(flags)
		self.readfile = fp
		self.writefile = fp
		self.local_path = local_path
		self.rel_path = os.path.normpath(rel_path.strip('/\\'))
		self.report_callable = report_callable
		self.stat_object = stat_object

	def close(self):
		super().close()
		if self.report_callable:
			self.report_callable(self.local_path, self.rel_path)

	def chattr(self, attr):
		_log.debug("chattr: %r", attr)
		return paramiko.SFTP_PERMISSION_DENIED

	def stat(self):
		if self.stat_object:
			return self.stat_object
		stat_obj = os.stat(self.local_path)
		basename = os.path.basename(self.local_path)
		return paramiko.SFTPAttributes.from_stat(stat_obj, basename)


class SFTPServerImpl(paramiko.SFTPServerInterface):
	__slots__ = (
			'u',
			'user_folder_path',
			'v_rev_filepath',
			'v_rev_content',
			'v_rev_stat',
			'_remote_location',
			'_process_callable',
	)

	def __init__(self, server: ServerImpl, transport: paramiko.Transport, *args, **kwargs):
		super().__init__(server, *args, **kwargs)
		self.u, _remote_username = server.lookup_user_via_transport(transport)
		self.user_folder_path = os.path.abspath(os.path.join(server.base_folder_path, self.u.username)) if self.u else '/dev/null'
		self.v_rev_filepath = '/' + server.v_rev_filename
		self.v_rev_content = server.v_rev_content
		self.v_rev_stat = server.v_rev_stat
		self._remote_location = _remote_location_from_transport(transport)
		self._process_callable = server.process_callable

	def _local_path(self, remote_path: str) -> str:
		p = os.path.abspath(os.path.join(self.user_folder_path, remote_path.strip('/\\')))
		if p.startswith(self.user_folder_path):
			return p
		_log.warning('escaped local path (user: %r): %r', self.user_folder_path, p)
		return self.user_folder_path

	def _do_report(self, local_path, rel_path):
		try:
			self._process_callable(self.u.username, self._remote_location, local_path, rel_path)
		except Exception:
			_log.exception("caught exception on invoking received file processor: user=%r, remote=%r, local_path=%r, rel_path=%r", self.u.username,
							self._remote_location, local_path, rel_path)

	def chattr(self, path, attr):
		_log.debug("chattr: %r, %r", path, attr)
		return paramiko.SFTP_PERMISSION_DENIED

	def list_folder(self, path):
		if not self.u:
			return paramiko.SFTP_PERMISSION_DENIED
		_log.debug("list_folder: %r", path)
		local_path = self._local_path(path)
		result = []
		aux = os.listdir(local_path)
		for n in aux:
			f_path = os.path.join(local_path, n)
			stat_obj = os.lstat(f_path)
			sftpattr = paramiko.SFTPAttributes.from_stat(stat_obj, n)
			result.append(sftpattr)
		if path == '/':
			result.append(self.v_rev_stat)
		return result

	def lstat(self, path):
		if not self.u:
			return paramiko.SFTP_PERMISSION_DENIED
		_log.debug("lstat: %r", path)
		if path == self.v_rev_filepath:
			return self.v_rev_stat
		local_path = self._local_path(path)
		try:
			stat_obj = os.lstat(local_path)
		except FileNotFoundError:
			return paramiko.SFTP_NO_SUCH_FILE
		return paramiko.SFTPAttributes.from_stat(stat_obj)

	def mkdir(self, path, attr):
		if not self.u:
			return paramiko.SFTP_PERMISSION_DENIED
		_log.debug("mkdir: %r, %r", path, attr)
		if path in ('/', self.v_rev_filepath):
			return paramiko.SFTP_PERMISSION_DENIED
		local_path = self._local_path(path)
		if local_path == self.user_folder_path:
			return paramiko.SFTP_PERMISSION_DENIED
		os.makedirs(local_path, 0o755, exist_ok=True)
		return paramiko.SFTP_OK

	def open(self, path, flags, attr):
		if not self.u:
			return paramiko.SFTP_PERMISSION_DENIED
		_log.debug("open: %r, %r, %r (%r)", path, flags, attr, self.user_folder_path)
		if path == self.v_rev_filepath:
			return SFTPHandle(io.BytesIO(self.v_rev_content), flags, path, path, None, self.v_rev_stat)
		local_path = self._local_path(path)
		fp = open(local_path, open_mode_from_flags(flags), opener=_fp_opener)  # pylint: disable=consider-using-with
		report_callable = self._do_report if ((flags & os.O_WRONLY) | (flags & os.O_RDWR)) else None
		return SFTPHandle(fp, flags, local_path, path, report_callable)

	def posix_rename(self, oldpath, newpath):
		if not self.u:
			return paramiko.SFTP_PERMISSION_DENIED
		_log.debug("posix_rename: %r, %r", oldpath, newpath)
		local_oldpath = self._local_path(oldpath)
		local_newpath = self._local_path(newpath)
		if self.user_folder_path in (local_oldpath, local_newpath):
			return paramiko.SFTP_PERMISSION_DENIED
		os.replace(local_oldpath, local_newpath)
		return paramiko.SFTP_OK

	def readlink(self, path):
		if not self.u:
			return paramiko.SFTP_PERMISSION_DENIED
		_log.debug("readlink: %r", path)
		local_path = self._local_path(path)
		realpath = os.path.realpath(local_path)
		if not realpath.startswith(self.user_folder_path):
			return paramiko.SFTP_NO_SUCH_FILE
		relatedpath = realpath[len(self.user_folder_path):]
		return relatedpath

	def remove(self, path):
		if not self.u:
			return paramiko.SFTP_PERMISSION_DENIED
		_log.debug("remove: %r", path)
		local_path = self._local_path(path)
		if local_path == self.user_folder_path:
			return paramiko.SFTP_PERMISSION_DENIED
		os.remove(local_path)
		return paramiko.SFTP_OK

	def rename(self, oldpath, newpath):
		if not self.u:
			return paramiko.SFTP_PERMISSION_DENIED
		_log.debug("rename: %r, %r", oldpath, newpath)
		local_oldpath = self._local_path(oldpath)
		local_newpath = self._local_path(newpath)
		if self.user_folder_path in (local_oldpath, local_newpath):
			return paramiko.SFTP_PERMISSION_DENIED
		shutil.move(local_oldpath, local_newpath)
		return paramiko.SFTP_OK

	def rmdir(self, path):
		if not self.u:
			return paramiko.SFTP_PERMISSION_DENIED
		_log.debug("rmdir: %r", path)
		local_path = self._local_path(path)
		if local_path == self.user_folder_path:
			return paramiko.SFTP_PERMISSION_DENIED
		os.rmdir(local_path)
		return paramiko.SFTP_OK

	def stat(self, path):
		if not self.u:
			return paramiko.SFTP_PERMISSION_DENIED
		_log.debug("stat: %r", path)
		if path == self.v_rev_filepath:
			return self.v_rev_stat
		local_path = self._local_path(path)
		stat_obj = os.stat(local_path)
		basename = os.path.basename(path)
		if not basename:
			basename = None
		return paramiko.SFTPAttributes.from_stat(stat_obj, basename)

	def symlink(self, target_path, path):
		if not self.u:
			return paramiko.SFTP_PERMISSION_DENIED
		_log.debug("symlink: %r, %r", target_path, path)
		local_target_path = self._local_path(target_path)
		local_sym_path = self._local_path(path)
		if self.user_folder_path in (local_target_path, local_sym_path):
			return paramiko.SFTP_PERMISSION_DENIED
		os.symlink(local_target_path, local_sym_path)
		return paramiko.SFTP_OK


class _ReportCallableWrap:
	__slots__ = (
			'u',
			'remote_location',
			'process_callable',
	)

	def __init__(self, channel: paramiko.Channel, u: User, process_callable: Callable[[str, str, str, str], None]) -> None:
		self.u = u
		self.remote_location = _remote_location_from_transport(channel.transport)
		self.process_callable = process_callable

	def __call__(self, local_path: str, rel_path: str) -> None:
		try:
			self.process_callable(self.u.username, self.remote_location, local_path, rel_path)
		except Exception:
			_log.exception("invoke process callable (%r, %r, %r, %r)", self.u.username, self.remote_location, local_path, rel_path)


class ServerImpl(paramiko.ServerInterface):
	__slots__ = (
			'base_folder_path',
			'users',
			'process_callable',
			'v_rev_filename',
			'v_rev_content',
			'v_rev_stat',
			'binpath_rsync',
			'_users_lck',
	)

	# pylint: disable=too-many-arguments
	def __init__(self,
					base_folder_path: str,
					users: Iterable[User],
					process_callable: Callable[[str, str, str, str], None],
					v_rev_filename: str = DEFAULT_REV_FILENAME,
					v_rev_content: str = DEFAULT_REV_CONTENT,
					binpath_rsync: Optional[str] = None) -> None:
		super().__init__()
		self.base_folder_path = base_folder_path
		self.users = make_users_map(users)
		self.process_callable = process_callable
		self.v_rev_filename = v_rev_filename
		self.v_rev_content = (v_rev_content.strip() + "\n").encode('utf-8', 'ignore')
		self.v_rev_stat = make_sftp_attr(v_rev_filename, self.v_rev_content)
		self.binpath_rsync = binpath_rsync
		self._users_lck = _thread.allocate_lock()

	def update_users(self, users: Iterable[User]):
		rec = make_users_map(users)
		with self._users_lck:
			to_drop = []
			for k in self.users:
				if k not in rec:
					to_drop.append(k)
			for k in to_drop:
				self.users.pop(k, None)
			for k, u in rec.items():
				self.users[k] = u

	def lookup_user_via_name(self, username: str) -> Optional[User]:
		if not username:
			return None
		with self._users_lck:
			return self.users.get(username)

	def lookup_user_via_transport(self, transport: paramiko.Transport) -> Tuple[Optional[User], str]:
		try:
			remote_username = transport.auth_handler.get_username()
		except Exception:
			_log.exception("cannot reach username for user lookup")
			return (None, '?')
		return (self.lookup_user_via_name(remote_username), remote_username)

	def _run_command(self, channel: paramiko.Channel, command: bytes) -> bool:
		u, remote_username = self.lookup_user_via_transport(channel.transport)
		if not u:
			_log.warning("cannot reach user [%r] for command: %r", remote_username, command)
			return False
		cmdpart = list(filter(None, command.decode(encoding='utf-8').split(' ')))
		if len(cmdpart) < 2:
			raise Exception('low argument count', cmdpart)
		cmdtarget = cmdpart[0]
		_log.info("run: [%s] as [%r]", cmdtarget, remote_username)
		user_folder_path = os.path.abspath(os.path.join(self.base_folder_path, u.username))
		report_callable = _ReportCallableWrap(channel, u, self.process_callable)
		if cmdtarget in ('rsync', '/bin/rsync', '/usr/bin/rsync'):
			if self.binpath_rsync:
				_thread.start_new_thread(run_rsync_exec, (user_folder_path, report_callable, channel, cmdpart))
			else:
				_log.warning("requested rsync but rsync executable path is not configurated")
				return False
		elif cmdtarget in ('scp', '/bin/scp', '/usr/bin/scp'):
			_thread.start_new_thread(run_scp_sink, (user_folder_path, report_callable, channel, cmdpart))
		else:
			raise Exception('unknown command', cmdtarget, command)
		return True

	def check_auth_password(self, username, password):
		# _log.debug("check_auth_password: %r, %r", username, password)
		u = self.lookup_user_via_name(username)
		if not u:
			return paramiko.AUTH_FAILED
		if not u.check_credential(password):
			return paramiko.AUTH_FAILED
		return paramiko.AUTH_SUCCESSFUL

	def check_auth_publickey(self, username, key):
		# _log.debug("check_auth_publickey: %r, %r", username, key)
		u = self.lookup_user_via_name(username)
		if not u:
			return paramiko.AUTH_FAILED
		key_type = key.get_name()
		key_b64 = key.get_base64()
		matched_key = u.check_ssh_pkey(key_type, key_b64)
		if not matched_key:
			return paramiko.AUTH_FAILED
		return paramiko.AUTH_SUCCESSFUL

	def check_channel_request(self, kind, chanid):
		_log.debug("check_channel_request: %r, %r", kind, chanid)
		return paramiko.OPEN_SUCCEEDED

	def get_allowed_auths(self, username):
		_log.debug("get_allowed_auths: %r", username)
		return "password,publickey"

	def check_channel_exec_request(self, channel, command):
		_log.debug("check_channel_exec_request: %r", command)
		# ref: https://github.com/carletes/mock-ssh-server/blob/master/mockssh/server.py
		if not channel.transport:
			return False
		try:
			return self._run_command(channel, command)
		except Exception:
			_log.exception('cannot invoke command')
		return False


class SSHLinkHandler(socketserver.BaseRequestHandler):
	def handle(self):
		_log.info("connected: %r", self.client_address)
		transport = paramiko.Transport(self.request)
		transport.add_server_key(self.server.host_pkey)
		transport.set_subsystem_handler('sftp', paramiko.SFTPServer, SFTPServerImpl, transport)
		transport.start_server(server=self.server.server_impl)
		ch = transport.accept()
		_log.debug("channel open: %r", ch)
		while transport.is_active():
			time.sleep(1)
		_log.info("disconnecting: %r", self.client_address)


class SFTPServer(socketserver.ForkingTCPServer):
	__slots__ = (
			'host_pkey',
			'server_impl',
	)

	# pylint: disable=too-many-arguments
	def __init__(self,
					server_host: str,
					server_port: int,
					key_file_path: str,
					key_bits: int,
					base_folder_path: str,
					users: Iterable[User],
					process_callable: Callable[[str, str, str, str], None],
					v_rev_filename: str = DEFAULT_REV_FILENAME,
					v_rev_content: str = DEFAULT_REV_CONTENT,
					binpath_rsync: Optional[str] = None) -> None:
		super().__init__((server_host, server_port), SSHLinkHandler)
		try:
			self.host_pkey = paramiko.RSAKey.from_private_key_file(key_file_path)
		except FileNotFoundError:
			self.host_pkey = paramiko.RSAKey.generate(key_bits)
			self.host_pkey.write_private_key_file(key_file_path)
		self.server_impl = ServerImpl(base_folder_path, users, process_callable, v_rev_filename, v_rev_content, binpath_rsync)

	def update_users(self, users: Iterable[User]):
		self.server_impl.update_users(users)
